@@ -2,33 +2,35 @@
   "Small clojure shim on top of the Kafka client API
    See https://github.com/pyr/kinsky for example usage."
   (:require [clojure.edn           :as edn]
-            [cheshire.core         :as json])
-  (:import java.util.Collection
-           java.util.Map
-           java.util.concurrent.TimeUnit
-           java.util.regex.Pattern
-           org.apache.kafka.clients.consumer.ConsumerRebalanceListener
-           org.apache.kafka.clients.consumer.ConsumerRecord
-           org.apache.kafka.clients.consumer.ConsumerRecords
-           org.apache.kafka.clients.consumer.KafkaConsumer
-           org.apache.kafka.clients.consumer.OffsetAndMetadata
-           org.apache.kafka.clients.producer.KafkaProducer
-           org.apache.kafka.clients.producer.ProducerRecord
-           org.apache.kafka.common.Node
-           org.apache.kafka.common.PartitionInfo
-           org.apache.kafka.common.TopicPartition
-           org.apache.kafka.common.errors.WakeupException
-           org.apache.kafka.common.serialization.Deserializer
-           org.apache.kafka.common.serialization.Serializer
-           org.apache.kafka.common.serialization.StringDeserializer
-           org.apache.kafka.common.serialization.StringSerializer))
+            [jsonista.core         :as json])
+  (:import (java.util Collection)
+           (java.util Map)
+           (java.util.concurrent TimeUnit)
+           (java.util.regex Pattern)
+           (org.apache.kafka.clients.consumer ConsumerRebalanceListener
+                                              ConsumerRecord
+                                              ConsumerRecords
+                                              KafkaConsumer
+                                              OffsetAndMetadata)
+           (org.apache.kafka.clients.producer KafkaProducer
+                                              ProducerRecord)
+           (org.apache.kafka.common Node
+                                    PartitionInfo
+                                    TopicPartition)
+           (org.apache.kafka.common.errors WakeupException)
+           (org.apache.kafka.common.serialization Deserializer
+                                                  Serializer
+                                                  StringDeserializer
+                                                  StringSerializer)
+           (org.apache.kafka.common.header Header
+                                           Headers)))
 
 (defprotocol MetadataDriver
   "Common properties for all drivers"
   (partitions-for [this topic]
     "Retrieve partition ownership information for a topic.
      The result is a data representation of a
-     [PartitionInfo](http://kafka.apache.org/090/javadoc/org/apache/kafka/common/PartitionInfo.html)
+     [PartitionInfo](https://kafka.apache.org/24/javadoc/org/apache/kafka/common/PartitionInfo.html)
      list.
      The structure for a partition info map is:
 
@@ -50,23 +52,31 @@
           :by-partition {[\"t\" 0] [{:key       \"k0\"
                                      :offset    1
                                      :partition 0
+                                     :timestamp 1233524522
                                      :topic     \"t\"
-                                     :value     \"v0\"}]
+                                     :value     \"v0\"
+                                     :headers   {:h1 [\"1\" \"2\"]}}]
                          [\"t\" 1] [{:key       \"k1\"
                                      :offset    1
                                      :partition 1
+                                     :timestamp 1233524527
                                      :topic     \"t\"
-                                     :value     \"v1\"}]}
+                                     :value     \"v1\"}
+                                     :headers   {:h1 [\"1\"]}]}
           :by-topic      {\"t\" [{:key       \"k0\"
                                   :offset    1
                                   :partition 0
+                                  :timestamp 1233524522
                                   :topic     \"t\"
-                                  :value     \"v0\"}
+                                  :value     \"v0\"
+                                  :headers   {:h1 [\"1\" \"2\"]}}
                                  {:key       \"k1\"
                                   :offset    1
                                   :partition 1
+                                  :timestamp 1233524527
                                   :topic     \"t\"
-                                  :value     \"v1\"}]}}")
+                                  :value     \"v1\"
+                                  :headers   {:h1 [\"1\"]}}]}}")
   (stop!          [this] [this timeout]
     "Stop consumption.")
   (pause!         [this] [this topic-partitions]
@@ -85,9 +95,9 @@
 
      The optional listener argument is either a callback
      function or an implementation of
-     [ConsumerRebalanceListener](http://kafka.apache.org/090/javadoc/org/apache/kafka/clients/consumer/ConsumerRebalanceListener.html).
+     [ConsumerRebalanceListener](https://kafka.apache.org/24/javadoc/org/apache/kafka/clients/consumer/ConsumerRebalanceListener.html).
 
-     When a function is supplied, it will be called on relance
+     When a function is supplied, it will be called on rebalance
      events with a map representing the event, see
      [kinsky.client/rebalance-listener](#var-rebalance-listener)
      for details on the map format.")
@@ -113,14 +123,17 @@
 
 (defprotocol ProducerDriver
   "Driver interface for producers"
-  (send!          [this record] [this topic k v]
+  (send!          [this record] [this topic k v] [this topic k v headers]
     "Produce a record on a topic.
      When using the single arity version, a map
      with the following possible keys is expected:
-     `:key`, `:topic`, `:partition`, and `:value`.
+     `:key`, `:topic`, `:partition`, `:headers`, `:timestamp` and `:value`.
      ")
   (flush!         [this]
-    "Ensure that produced messages are flushed."))
+    "Ensure that produced messages are flushed.")
+  (init-transactions! [this])
+  (begin-transaction! [this])
+  (commit-transaction! [this]))
 
 (defprotocol GenericDriver
   (close!         [this] [this timeout]
@@ -139,23 +152,27 @@
 
 (defn edn-serializer
   "Serialize as EDN."
+  ^Serializer
   []
   (serializer
    (fn [_ payload] (some-> payload pr-str .getBytes))))
 
 (defn json-serializer
-  "Serialize as JSON through cheshire."
+  "Serialize as JSON through jsonista."
+  ^Serializer
   []
   (serializer
-   (fn [_ payload] (some-> payload json/generate-string .getBytes))))
+   (fn [_ payload] (some-> payload json/write-value-as-bytes))))
 
 (defn keyword-serializer
   "Serialize keywords to strings, useful for keys."
+  ^Serializer
   []
   (serializer (fn [_ k] (some-> k name .getBytes))))
 
 (defn string-serializer
   "Kafka's own string serializer."
+  ^StringSerializer
   []
   (StringSerializer.))
 
@@ -172,27 +189,37 @@
 
 (defn edn-deserializer
   "Deserialize EDN."
-  []
-  (deserializer
-   (fn [_ #^"[B" payload]
-     (when payload
-       (edn/read-string (String. payload "UTF-8"))))))
+  (^Deserializer []
+   (deserializer
+    (fn [_ #^"[B" payload]
+      (when payload
+        (edn/read-string (String. payload "UTF-8"))))))
+  (^Deserializer [reader-opts]
+   (deserializer
+     (fn [_ #^"[B" payload]
+       (when payload
+         (edn/read-string reader-opts (String. payload "UTF-8")))))))
 
 (defn json-deserializer
   "Deserialize JSON."
+  ^Deserializer
   []
-  (deserializer
-   (fn [_ #^"[B" payload]
-     (when payload
-       (json/parse-string (String. payload "UTF-8") true)))))
+  (let [mapper (json/object-mapper {:encode-key-fn name
+                                    :decode-key-fn keyword})]
+    (deserializer
+     (fn [_ #^"[B" payload]
+       (when payload
+         (json/read-value payload mapper))))))
 
 (defn keyword-deserializer
   "Deserialize a string and then keywordize it."
+  ^Deserializer
   []
   (deserializer (fn [_ #^"[B" k] (when k (keyword (String. k "UTF-8"))))))
 
 (defn string-deserializer
   "Kafka's own string deserializer"
+  ^StringDeserializer
   []
   (StringDeserializer.))
 
@@ -208,7 +235,8 @@
    :string  string-serializer
    :json    json-serializer})
 
-(defn ^Deserializer ->deserializer
+(defn ->deserializer
+  ^Deserializer
   [x]
   (cond
     (keyword? x) (if-let [f (deserializers x)]
@@ -217,27 +245,34 @@
     (fn? x)      (x)
     :else        x))
 
-(defn ^Serializer ->serializer
+(defn ->serializer
+  ^Serializer
   [x]
   (cond
     (keyword? x) (if-let [f (serializers x)]
                    (f)
                    (throw (ex-info "unknown serializer alias" {})))
-    (fn? x)      (x)
+    (fn? x)     (x)
     :else        x))
 
-(defn ^Map opts->props
-  "Kakfa configs are now maps of strings to strings. Morph
-   an arbitrary clojure map into this representation."
+(defn opts->props
+  "Kakfa configs are now maps of strings to strings. Morphs an arbitrary
+  clojure map into this representation.  Make sure we don't pass
+  options that are meant for the driver to concrete Consumers/Producers"
+  ^Map
   [opts]
-  (into {} (for [[k v] opts] [(name k) (str v)])))
+  (into {}
+        (comp
+         (filter (fn [[k _]] (not (qualified-keyword? k))))
+         (map (fn [[k v]] [(name k) (str v)])))
+        opts))
 
-(defn ^ConsumerRebalanceListener rebalance-listener
+(defn rebalance-listener
   "Wrap a callback to yield an instance of a Kafka ConsumerRebalanceListener.
    The callback is a function of one argument, a map containing the following
    keys: :event, :topic and :partition. :event will be either :assigned or
    :revoked."
-  [callback]
+  ^ConsumerRebalanceListener [callback]
   (if (instance? ConsumerRebalanceListener callback)
     callback
     (let [->part  (fn [^TopicPartition p] {:topic (.topic p) :partition (.partition p)})
@@ -253,7 +288,7 @@
 
 (defn ->topic-partition
   "Yield a TopicPartition from a clojure map."
-  [{:keys [topic partition]}]
+  ^TopicPartition [{:keys [topic partition]}]
   (TopicPartition. (name topic) (int partition)))
 
 (defn ->offset-metadata
@@ -287,7 +322,7 @@
   "A transducer to explode grouped records into individual
    entities.
 
-   When sucessful, the output of kinsky.client/poll! takes the
+   When successful, the output of kinsky.client/poll! takes the
    form:
 
        {:partitions   [[\"t\" 0] [\"t\" 1]]
@@ -324,31 +359,75 @@
   "
   (comp (map :by-partition) (mapcat vals) cat))
 
+(def value-meta-xform
+  "A transducer which acts like `record-xform` but directly yields values,
+   with record metadata stored in the map metadata"
+  (comp record-xform
+        (map (fn [{:keys [value] :as record}]
+               (with-meta value
+                 (select-keys record [:key :offset :partition :topic]))))))
+
+(def value-xform
+  "A transducer which acts like `record-xform` but directly yields values,
+   losing metadata"
+  (comp record-xform (map :value)))
+
+(defn headers->map
+  "Build a clojure map out of Kafka Headers from the returned RecordHeaders object. As the Kafka Headers can have
+  duplicates, the map returned merges any such cases: {:h1 [\"123\" \"456\"] :h2 [\"Hello\"]}"
+  [^Headers headers]
+  (apply merge-with into {} (map (fn [^Header header]
+                                   {(keyword (.key header))
+                                    [(slurp (.value header) :encoding "UTF-8")]})
+                                 headers)))
+
 (defn cr->data
   "Yield a clojure representation of a consumer record"
   [^ConsumerRecord cr]
   {:key       (.key cr)
    :offset    (.offset cr)
    :partition (.partition cr)
+   :timestamp (.timestamp cr)
    :topic     (.topic cr)
-   :value     (.value cr)})
+   :value     (.value cr)
+   :headers   (headers->map (.headers cr))})
+
+(defn crs->eduction
+  "Returns consumer records as clojure.lang.Eduction to be used in
+  potential reducible context"
+  [^ConsumerRecords crs]
+  (eduction (map cr->data) crs))
+
+(defn crs-for-topic->eduction
+  "Returns an Eduction for records by topics"
+  [^ConsumerRecords crs ^String topic]
+  (eduction (map cr->data) (.records crs topic)))
+
+(defn crs-for-topic+partition->eduction
+  "Returns an Eduction for records for TopicPartition"
+  [^ConsumerRecords crs topic partition]
+  (let [tp (TopicPartition. topic
+                            (int partition))]
+    (eduction (map cr->data) (.records crs tp))))
 
 (defn consumer-records->data
-  "Yield the clojure representation of topic"
+  "Yield the default Clojure representation of topic"
   [^ConsumerRecords crs]
-  (let [->d  (fn [^TopicPartition p] [(.topic p) (.partition p)])
-        ps   (.partitions crs)
-        ts   (set (for [^TopicPartition p ps] (.topic p)))
-        by-p (into {} (for [^TopicPartition p ps] [(->d p) (mapv cr->data (.records crs p))]))
-        by-t (into {} (for [^String t ts] [t (mapv cr->data (.records crs t))]))]
-    {:partitions   (vec (for [^TopicPartition p ps] [(.topic p) (.partition p)]))
+  (let [ps (.partitions crs)
+        ts   (into #{}
+                   (map (fn [^TopicPartition p]
+                          (.topic p)))
+                   ps)
+        edc (crs->eduction crs)]
+    {:partitions   ps
      :topics       ts
      :count        (.count crs)
-     :by-topic     by-t
-     :by-partition by-p}))
+     :by-topic     (group-by :topic edc)
+     :by-partition (group-by (juxt :topic :partition) edc)}))
 
-(defn ^Collection ->topics
+(defn ->topics
   "Yield a valid object for subscription"
+  ^Collection
   [topics]
   (cond
     (keyword? topics)             [(name topics)]
@@ -366,20 +445,28 @@
    - [ConsumerDriver](#var-ConsumerDriver)
    - [MetadataDriver](#var-MetadataDriver)
    - `clojure.lang.IDeref`: `deref` to access underlying
-     [KafkaConsumer](http://kafka.apache.org/090/javadoc/org/apache/kafka/clients/producer/KafkaConsumer.html)
-     instance."
+     [KafkaConsumer](https://kafka.apache.org/24/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html)
+     instance.
+
+  The consumer-driver can also take options:
+
+  * `consumer-decoder-fn`: a function that will potentially transform
+  the ConsumerRecords returned by kafka. By default it will use
+  `consumer-records->data`"
   ([^KafkaConsumer consumer]
    (consumer->driver consumer nil))
-  ([^KafkaConsumer consumer run-signal]
+  ([^KafkaConsumer consumer
+    {::keys [run-fn consumer-decoder-fn]
+     :or {consumer-decoder-fn consumer-records->data}}]
    (reify
      ConsumerDriver
      (poll! [this timeout]
-       (consumer-records->data (.poll consumer timeout)))
+       (consumer-decoder-fn (.poll consumer (java.time.Duration/ofMillis timeout))))
      (stop! [this]
        (stop! this 0))
      (stop! [this timeout]
-       (when run-signal
-         (run-signal))
+       (when (fn? run-fn)
+         (run-fn))
        (.wakeup consumer))
      (pause! [this topic-partitions]
        (.pause consumer
@@ -412,25 +499,27 @@
      (commit! [this]
        (.commitSync consumer))
      (commit! [this topic-offsets]
-       (.commitSync consumer
-                    (->> topic-offsets
-                         (map (juxt ->topic-partition ->offset-metadata))
-                         (reduce merge {}))))
+       (let [offsets (->> topic-offsets
+                          (map (juxt ->topic-partition ->offset-metadata))
+                          (reduce merge {}))]
+         (.commitSync consumer ^Map offsets)))
     (seek! [this topic-partition offset]
-        (.seek consumer (->topic-partition topic-partition) offset))
+      (.seek consumer
+             (->topic-partition topic-partition)
+             (long offset)))
     (position! [this topic-partition]
       (.position consumer (->topic-partition topic-partition)))
-     (subscription [this]
-       (.subscription consumer))
+    (subscription [this]
+      (.subscription consumer))
     GenericDriver
     (close! [this]
       (.close consumer))
-     MetadataDriver
-     (partitions-for [this topic]
-       (mapv partition-info->data (.partitionsFor consumer topic)))
-     clojure.lang.IDeref
-     (deref [this]
-       consumer))))
+    MetadataDriver
+    (partitions-for [this topic]
+      (mapv partition-info->data (.partitionsFor consumer topic)))
+    clojure.lang.IDeref
+    (deref [this]
+      consumer))))
 
 (defn safe-poll!
   "Implementation of poll which disregards wake-up exceptions"
@@ -439,26 +528,43 @@
     (poll! consumer timeout)
     (catch WakeupException _)))
 
-(defn ->record
+(defn ->header
+  "Build a Kafka header from a key and a value"
+  (^Header
+   [[k v]]
+   (->header k v))
+  (^Header
+   [k v]
+   (reify
+     Header
+     (key [_] (if (keyword? k) (name k) (str k)))
+     (value [_] (byte-array (map byte (str v))))
+     (toString [_] (str "RecordHeader(key = " k ", value = " v ")")))))
+
+(defn ->headers
+  "Build Kafka headers from a clojure map."
+  [headers]
+  (map ->header headers))
+
+(defn ^ProducerRecord ->record
   "Build a producer record from a clojure map. Leave ProducerRecord instances
-   untouched."
+  untouched."
   [payload]
   (if (instance? ProducerRecord payload)
     payload
-    (let [{:keys [partition key value]} payload
-          topic                         (some-> payload :topic name)]
-      (cond
-        (nil? topic)
-        (throw (ex-info "Need a topic to send to" {:payload payload}))
+    (let [{:keys [partition key value headers timestamp]} payload
+          topic (some-> payload :topic name)]
+      (when (nil? topic)
+        (throw (ex-info "Need a topic to send to" {:payload payload})))
 
-        (and key partition)
-        (ProducerRecord. (str topic) (int partition) key value)
+      (when (and headers (nil? timestamp) (nil? partition))
+        (throw (ex-info "Need a partition and/or timestamp if you want to add headers" {:payload payload})))
 
-        key
-        (ProducerRecord. (str topic) key value)
-
-        :else
-        (ProducerRecord. (str topic) value)))))
+      (ProducerRecord. (str topic)
+                       (when partition (int partition))
+                       (when timestamp (long timestamp))
+                       key value
+                       (->headers headers)))))
 
 (defn producer->driver
   "Yield a driver from a Kafka Producer.
@@ -467,7 +573,7 @@
    - [ProducerDriver](#var-ProducerDriver)
    - [MetadataDriver](#var-MetadataDriver)
    - `clojure.lang.IDeref`: `deref` to access underlying
-     [KafkaProducer](http://kafka.apache.org/090/javadoc/org/apache/kafka/clients/producer/KafkaProducer.html)
+     [KafkaProducer](https://kafka.apache.org/24/javadoc/org/apache/kafka/clients/producer/KafkaProducer.html)
      instance."
   [^KafkaProducer producer]
   (reify
@@ -476,15 +582,24 @@
       (.close producer))
     (close! [this timeout]
       (if (nil? timeout)
-        (.close timeout)
+        (.close producer)
         (.close producer (long timeout) TimeUnit/MILLISECONDS)))
     ProducerDriver
     (send! [this record]
       (.send producer (->record record)))
     (send! [this topic k v]
       (.send producer (->record {:key k :value v :topic topic})))
+    (send! [this topic k v headers]
+      "Defaults partition and timestamp to 0, if you need to set either, use the single arity version"
+      (.send producer (->record {:key k :value v :topic topic :headers headers})))
     (flush! [this]
       (.flush producer))
+    (init-transactions! [this]
+      (.initTransactions producer))
+    (begin-transaction! [this]
+      (.beginTransaction producer))
+    (commit-transaction! [this]
+      (.commitTransaction producer))
     MetadataDriver
     (partitions-for [this topic]
       (mapv partition-info->data (.partitionsFor producer topic)))
@@ -514,16 +629,18 @@
    If deserializers are provided, use them otherwise expect deserializer
    class name in the config map."
   ([config]
-   (consumer->driver (KafkaConsumer. (opts->props config))))
+   (consumer->driver (KafkaConsumer. (opts->props config))
+                     config))
   ([config callback]
    (consumer->driver (KafkaConsumer. (opts->props config))
-                     callback))
+                     (assoc config ::run-fn callback)))
   ([config kdeserializer vdeserializer]
    (consumer->driver (KafkaConsumer. (opts->props config)
                                      (->deserializer kdeserializer)
-                                     (->deserializer vdeserializer))))
+                                     (->deserializer vdeserializer))
+                     config))
   ([config callback kdeserializer vdeserializer]
    (consumer->driver (KafkaConsumer. (opts->props config)
                                      (->deserializer kdeserializer)
                                      (->deserializer vdeserializer))
-                     callback)))
+                     (assoc config ::run-fn callback))))
