@@ -6,6 +6,7 @@
   (:import (java.io File))
   (:require [riemann [boundary    :refer [boundary]]
                      [client      :refer [udp-client tcp-client multi-client]]
+                     [clickhouse  :refer [clickhouse]]
                      [cloudwatch  :refer [cloudwatch]]
                      [common      :as common :refer [event]]
                      [core        :as core]
@@ -17,6 +18,7 @@
                      [hipchat     :refer [hipchat]]
                      [index       :as index]
                      [influxdb    :refer [influxdb]]
+                     [influxdb2   :refer [influxdb2]]
                      [kafka       :as kafka :refer [kafka]]
                      [kairosdb    :refer [kairosdb]]
                      [keenio      :refer [keenio]]
