@@ -1,7 +1,9 @@
 (ns amazonica.core
   "Amazon AWS functions."
   (:use [clojure.algo.generic.functor :only (fmap)])
-  (:require [clojure.string :as str])
+  (:require [clojure.string :as str]
+            [clojure.java.io :as io]
+            [clojure.set :as set])
   (:import clojure.lang.Reflector
            [com.amazonaws
              AmazonServiceException
@@ -9,6 +11,7 @@
            [com.amazonaws.auth
              AWSCredentials
              AWSCredentialsProvider
+             AWSStaticCredentialsProvider
              BasicAWSCredentials
              BasicSessionCredentials
              DefaultAWSCredentialsProviderChain]
@@ -16,6 +19,8 @@
            [com.amazonaws.regions
              Region
              Regions]
+           com.amazonaws.client.builder.AwsClientBuilder
+           com.amazonaws.client.builder.AwsClientBuilder$EndpointConfiguration
            org.joda.time.DateTime
            org.joda.time.base.AbstractInstant
            java.io.File
@@ -23,7 +28,10 @@
            java.io.StringWriter
            java.lang.reflect.InvocationTargetException
            java.lang.reflect.ParameterizedType
+           java.lang.reflect.Method
            java.lang.reflect.Modifier
+           java.lang.reflect.Constructor
+           java.lang.reflect.Parameter
            java.math.BigDecimal
            java.math.BigInteger
            java.nio.ByteBuffer
@@ -71,7 +79,7 @@
 
 (defn stack->string
   "Converts a Java stacktrace to String representation."
-  [ex]
+  [^Throwable ex]
   (let [sw (StringWriter.)
         pw (PrintWriter. sw)
         _  (.printStackTrace ex pw)]
@@ -87,7 +95,7 @@
   :service-name
   :message
   :stack-trace"
-  [e]
+  [^AmazonServiceException e]
   {:error-code   (.getErrorCode e)
    :error-type   (.toString (.getErrorType e))
    :status-code  (.getStatusCode e)
@@ -134,44 +142,61 @@
   "Per invocation binding of client-config for ad-hoc
   service calls using alternate client configuration."
   [config & body]
-  `(binding [*client-config* config]
+  `(binding [*client-config* ~config]
      (do ~@body)))
 
-(declare new-instance)
+(defn- builder ^AwsClientBuilder [^Class clazz]
+  (let [^Method method (.getMethod clazz "builder" (make-array Class 0))]
+    (.invoke method clazz (make-array Object 0))))
+
+(declare set-fields)
+
+(defn- build-client [^Class clazz credentials configuration raw-creds options]
+  (let [builder (builder clazz)
+        _ (set-fields builder options)
+        builder (if credentials (.withCredentials builder credentials) builder)
+        builder (if configuration (.withClientConfiguration builder configuration) builder)
+        ^String endpoint (or (:endpoint raw-creds) (System/getenv "AWS_DEFAULT_REGION"))
+        builder (if endpoint
+                  (if (.startsWith endpoint "http")
+                      (.withEndpointConfiguration
+                        builder
+                        (AwsClientBuilder$EndpointConfiguration. endpoint nil))
+                      (.withRegion builder endpoint))
+                  builder)]
+    (.build builder)))
 
 (defn- create-client
-  [clazz credentials configuration]
-  (if (every? nil? [credentials configuration])
-    (new-instance (Class/forName clazz))
-    ; TransferManager is the only client to date that doesn't accept AWSCredentialsProviders
-    (if (= (.getSimpleName clazz) "TransferManager")
-        (invoke-constructor
-          "com.amazonaws.services.s3.transfer.TransferManager"
-          [(create-client (Class/forName "com.amazonaws.services.s3.AmazonS3Client")
-                          credentials
-                          configuration)])
-        (invoke-constructor (.getName clazz)
-                            (->> [credentials configuration]
-                                 (filter (comp not nil?))
-                                 vec)))))
+  [^Class clazz credentials configuration raw-creds options]
+  (if (= (.getSimpleName clazz) "TransferManager")
+    (invoke-constructor
+      "com.amazonaws.services.s3.transfer.TransferManager"
+      [(create-client (Class/forName "com.amazonaws.services.s3.AmazonS3Client")
+                      credentials
+                      configuration
+                      raw-creds {})])
+    (build-client clazz credentials configuration raw-creds options)))
 
-(defn get-credentials
+(defn get-credentials ^AWSCredentialsProvider
   [credentials]
   (cond
-    (or (instance? AWSCredentialsProvider credentials)
-        (instance? AWSCredentials credentials))
-    credentials
+    (instance? AWSCredentialsProvider credentials)
+      credentials
+    (instance? AWSCredentials credentials)
+      (AWSStaticCredentialsProvider. credentials)
     (and (associative? credentials)
          (contains? credentials :session-token))
-    (BasicSessionCredentials.
+    (AWSStaticCredentialsProvider.
+      (BasicSessionCredentials.
         (:access-key credentials)
         (:secret-key credentials)
-        (:session-token credentials))
+        (:session-token credentials)))
     (and (associative? credentials)
          (contains? credentials :access-key))
-    (BasicAWSCredentials.
+    (AWSStaticCredentialsProvider.
+      (BasicAWSCredentials.
         (:access-key credentials)
-        (:secret-key credentials))
+        (:secret-key credentials)))
     (and (associative? credentials)
          (contains? credentials :profile))
     (ProfileCredentialsProvider.
@@ -180,7 +205,7 @@
          (instance? AWSCredentialsProvider (:cred credentials)))
     (:cred credentials)
     :else
-    (DefaultAWSCredentialsProviderChain.)))
+    (DefaultAWSCredentialsProviderChain/getInstance)))
 
 (defn parse-args
   "Legacy support means credentials may or may not be passed
@@ -188,7 +213,8 @@
   [cred args]
   (if (and (instance? AWSCredentialsProvider (get-credentials cred))
            (not (instance? AWSCredentialsProvider cred))
-           (nil? (:endpoint cred)))
+           (nil? (:endpoint cred))
+           (nil? (:profile cred)))
     {:args (conj args cred)}
     {:args args :cred cred}))
 
@@ -199,8 +225,8 @@
   (when (associative? configuration)
     (create-bean ClientConfiguration configuration)))
 
-(defn- set-endpoint!
-  [client credentials]
+(defn- get-region ^Regions
+  [credentials]
   (when-let [endpoint (or (:endpoint credentials)
                           (System/getenv "AWS_DEFAULT_REGION"))]
     (if (contains? (fmap #(-> % str/upper-case (str/replace "_" ""))
@@ -210,27 +236,30 @@
         (try
           (->> (-> (str/upper-case endpoint)
                    (str/replace "-" "_"))
-               Regions/valueOf
-               Region/getRegion
-               (.setRegion client))
+               Regions/valueOf)
           (catch NoSuchMethodException e
-            (println e)))
-        (.setEndpoint client endpoint))))
+            (println e))))))
 
 (defn encryption-client*
   [encryption credentials configuration]
   (let [creds     (get-credentials credentials)
         config    (get-client-configuration configuration)
         key       (or (:secret-key encryption)
-                      (:key-pair encryption))
+                      (:key-pair encryption)
+                      (:kms-customer-master-key encryption))
         crypto    (invoke-constructor
                     "com.amazonaws.services.s3.model.CryptoConfiguration" [])
-        em        (invoke-constructor
-                    "com.amazonaws.services.s3.model.EncryptionMaterials"
-                    [key])
-        materials (invoke-constructor
+        _         (when (:kms-customer-master-key encryption)
+                    (.setAwsKmsRegion crypto (Region/getRegion (get-region {:endpoint (:region key)})) ))
+        em        (when-not (:kms-customer-master-key encryption)
+                    (invoke-constructor "com.amazonaws.services.s3.model.EncryptionMaterials"
+                    [key]))
+        materials (if (:kms-customer-master-key encryption)
+                    (invoke-constructor "com.amazonaws.services.s3.model.KMSEncryptionMaterialsProvider"
+                    [(:id key)])
+                    (invoke-constructor
                     "com.amazonaws.services.s3.model.StaticEncryptionMaterialsProvider"
-                    [em])
+                    [em]))
         _         (if-let [provider (:provider encryption)]
                     (.withCryptoProvider crypto provider))
         client    (if config
@@ -240,7 +269,6 @@
                       (invoke-constructor
                         "com.amazonaws.services.s3.AmazonS3EncryptionClient"
                         [creds materials crypto]))]
-    (set-endpoint! client credentials)
     client))
 
 (swap! client-config assoc :encryption-client-fn (memoize encryption-client*))
@@ -249,8 +277,7 @@
   [clazz credentials configuration]
   (let [aws-creds  (get-credentials credentials)
         aws-config (get-client-configuration configuration)
-        client     (create-client clazz aws-creds aws-config)]
-    (set-endpoint! client credentials)
+        client     (create-client clazz aws-creds aws-config credentials configuration)]
     client))
 
 (swap! client-config assoc :amazon-client-fn (memoize amazon-client*))
@@ -278,7 +305,7 @@
    #(str/replace % #"(?i)iscsi" "ISCSI")
    #(str/replace % #"(?i)openid" "OPENID")))
 
-(defn- keyword->camel
+(defn- ^String keyword->camel
   [kw]
   (let [n (name kw)
         m (.replace n "?" "")]
@@ -287,7 +314,7 @@
          str/join)))
 
 (defn- aws-package?
-  [clazz]
+  [^Class clazz]
   (->> (.getName clazz)
        (re-find #"com\.amazonaws\.services")
        some?))
@@ -296,8 +323,8 @@
   [date]
   (cond
     (instance? java.util.Date date) date
-    (instance? AbstractInstant date) (.toDate date)
-    (integer? date) (java.util.Date. date)
+    (instance? AbstractInstant date) (.toDate ^AbstractInstant date)
+    (integer? date) (java.util.Date. (long date))
     true (.. (SimpleDateFormat. @date-format)
            (parse (str date) (ParsePosition. 0)))))
 
@@ -306,17 +333,18 @@
   (if (instance? File file)
     file
     (if (string? file)
-      (File. file))))
+      (File. ^String file))))
 
 (defn to-enum
   "Case-insensitive resolution of Enum types by String."
-  [type value]
+  ^java.lang.Enum [^Class type value]
   (some
-    #(if (and
-           (not (nil? (.toString %))) ; some aws enums return nil!
-           (= (str/upper-case value)
-             (str/upper-case (.toString %))))
-      %)
+    (fn [^java.lang.Enum en]
+      (if (and
+            (not (nil? (.toString en))) ; some aws enums return nil!
+            (= (str/upper-case value)
+               (str/upper-case (.toString en))))
+        en))
     (-> type
         (.getDeclaredMethod "values" (make-array Class 0))
         (.invoke type (make-array Object 0)))))
@@ -360,7 +388,7 @@
   "Coerces the supplied stringvalue to the required type as
   defined by the AWS method signature. String or keyword
   conversion to Enum types (e.g. via valueOf()) is supported."
-  [value type]
+  [value ^Class type]
   (let [value (if (keyword? value) (name value) value)]
     (if-not (instance? type value)
       (if (= java.lang.Enum (.getSuperclass type))
@@ -385,7 +413,7 @@
     class-name))
 
 (defn- constructor-args
-  [ctor]
+  [^Constructor ctor]
   (let [types (.getParameterTypes ctor)]
     (if (= 0 (count types))
       (make-array Object 0)
@@ -396,17 +424,16 @@
 
 (defn best-constructor
   "Prefer no-arg ctor if one exists, else the first found."
-  [clazz]
+  ^Constructor [^Class clazz]
   {:pre [clazz]}
   (let [ctors (.getConstructors clazz)]
     (or
       (some
-        #(if (= 0 (count (.getParameterTypes %)))
-          %
-          nil)
+        (fn [^Constructor ctor]
+          (when (= 0 (count (.getParameterTypes ctor))) ctor))
         ctors)
       (first (sort-by
-               (fn [ctor]
+               (fn [^Constructor ctor]
                  (some aws-package? (.getParameterTypes ctor)))
                ctors)))))
 
@@ -424,19 +451,19 @@
   [depth param]
   (if (instance? ParameterizedType param)
       (let [f (partial unwind-types (inc depth))
-            types (-> param .getActualTypeArguments)
+            types (-> ^ParameterizedType param .getActualTypeArguments)
             t (-> types last f)]
         (if (and (instance? ParameterizedType (last types))
                  (.contains (-> types last str) "java.util")
                  (.contains (-> types last str) "java.lang"))
-          {:type [(-> types last .getRawType)]
+          {:type [(-> types ^ParameterizedType last .getRawType)]
            :depth depth}
           t))
       {:type [param]
        :depth depth}))
 
 (defn- paramter-types
-  [method]
+  [^Method method]
   (let [types (seq (.getGenericParameterTypes method))
         param (last types)
         rval  {:generic types}]
@@ -448,8 +475,8 @@
         rval)))
 
 (defn- normalized-name
-  [method-name]
-  (-> (name method-name)
+  [^String method-name]
+  (-> method-name
       (.replaceFirst
         (case (.substring method-name 0 3)
           "get" "get"
@@ -467,22 +494,24 @@
    in the DynamoDBV2Client.
    http://docs.aws.amazon.com/AWSJavaSDK/latest/javadoc/com/amazonaws/services/dynamodbv2/model/DeleteItemRequest.html#setKey(java.util.Map.Entry, java.util.Map.Entry)
    http://docs.aws.amazon.com/AWSJavaSDK/latest/javadoc/com/amazonaws/services/dynamodbv2/model/GetItemRequest.html#setKey(java.util.Map.Entry, java.util.Map.Entry)"
-  [method name value]
-  (let [args (.getParameterTypes method)]
+  [^Method method name value]
+  (let [args (.getParameterTypes method)
+        n (count args)
+        ^Class clazz (first args)]
     (and (= name (normalized-name (.getName method)))
          (if (empty? value)
-             (= 0 (count args))
-             (and (< 0 (count args))
-                  (or (= (count args) (count (flatten value)))
+             (= 0 n)
+             (and (< 0 n)
+                  (or (= n (count (flatten value)))
                       (and (coll? value)
-                           (= 1 (count args))
-                           (or (contains? @coercions (first args))
-                               (.isArray (first args))
-                               (and (.getPackage (first args))
+                           (= 1 n)
+                           (or (contains? @coercions clazz)
+                               (.isArray clazz)
+                               (and (.getPackage clazz)
                                     (.startsWith
-                                      (.getName (.getPackage (first args)))
+                                      (.getName (.getPackage clazz))
                                       "java.util")))))
-                  (not (and (= 2 (count args))
+                  (not (and (= 2 n)
                             (every? (partial = java.util.Map$Entry)
                                     args))))))))
 
@@ -496,7 +525,7 @@
     class-methods))
 
 (defn- find-methods
-  [pojo k & v]
+  [^Object pojo k & v]
   (-> (.getClass pojo)
       (.getMethods)
       (accessor-methods
@@ -513,18 +542,18 @@
         (java.util.HashMap.)
         (.putAll col))
     (set? col)
-      (java.util.HashSet. col)
+      (java.util.HashSet. ^java.util.Set col)
     (or (list? col) (vector? col))
-      (java.util.ArrayList. col)))
+      (java.util.ArrayList. ^java.util.Collection col)))
 
 (defn- invoke
-  [pojo method v]
+  [pojo ^Method method v]
   (.invoke method pojo
     (into-array
       ;; misbehaving S3Client mutates the coll
       (if (and (coll? v)
                (not (nil? *client-class*))
-               (= "AmazonS3Client" (.getSimpleName *client-class*)))
+               (= "AmazonS3Client" (.getSimpleName ^Class *client-class*)))
         (if (and (> (count (.getParameterTypes method)) 1)
                  (sequential? v))
           (to-java-coll v)
@@ -532,8 +561,6 @@
           [v])))
   true)
 
-(declare set-fields)
-
 (defn kw->str [k]
   (if (keyword? k) (name k) k))
 
@@ -548,23 +575,25 @@
 (defn- unmarshall
   "Transform Clojure data to the required Java objects."
   [types col]
-  (let [type (last (or (:actual types)
-                       (:generic types)))
+  (let [^Class type (last (or (:actual types)
+                              (:generic types)))
         pp   (partial populate types :actual)]
     (try
       (if (aws-package? type)
        (if (map? col)
          (if (contains? types :actual)
-           (if (< (:depth types) 3)
-             (apply assoc {}
-                    (interleave (fmap kw->str (apply vector (keys col)))
-                                (fmap pp (apply vector (vals col)))))
-             (apply assoc {}
-                    (interleave (fmap kw->str (apply vector (keys col)))
-                                [(fmap #(populate {:generic [type]}
-                                                  :generic
-                                                  %)
-                                       (first (apply vector (vals col))))])))
+           (if (empty? col)
+             {}  ; assoc doesn't like being called to do nothing
+             (if (< (:depth types) 3)
+               (apply assoc {}
+                      (interleave (fmap kw->str (apply vector (keys col)))
+                                  (fmap pp (apply vector (vals col)))))
+               (apply assoc {}
+                      (interleave (fmap kw->str (apply vector (keys col)))
+                                  [(fmap #(populate {:generic [type]}
+                                                    :generic
+                                                    %)
+                                         (first (apply vector (vals col))))]))))
            (populate types :generic col))
          (if (and (contains? types :actual)
                   (= (:depth types) 3))
@@ -621,7 +650,7 @@
    bean passed as the argument to a method call on the
    Amazon*Client class. (Note that we assume all AWS
    service calls take at most a single argument.)"
-  [method args]
+  [^Method method args]
   (let [clazz (first (.getParameterTypes method))]
     (if (contains? @coercions clazz)
         (coerce-value (into {} args) clazz)
@@ -639,7 +668,7 @@
   (marshall [obj]))
 
 (defn- getter?
-  [method]
+  [^Method method]
   (let [name (.getName method)
         type (.getName (.getReturnType method))]
     (or (and
@@ -651,19 +680,20 @@
 
 (defn accessors
   "Returns a vector of getters or setters for the class."
-  [clazz getters?]
+  [^Class clazz getters?]
   (reduce
-    #(if (or
-           (and getters? (getter? %2))
-           (and (not getters?)
-                (.startsWith (.getName %2) "set")))
-      (conj % %2)
-      %)
+    (fn [acc ^Method m]
+      (if (or
+            (and getters? (getter? m))
+            (and (not getters?)
+                 (.startsWith (.getName m) "set")))
+        (conj acc m)
+        acc))
     []
     (.getDeclaredMethods clazz)))
 
 (defn- prop->name
-  [method]
+  [^Method method]
   (let [name (.getName method)]
     (if (.startsWith name "is")
       (str (.substring name 2) "?")
@@ -675,7 +705,7 @@
   [obj]
   (let [no-arg (make-array Object 0)]
     (into {}
-      (for [m (accessors (class obj) true)]
+      (for [^Method m (accessors (class obj) true)]
         (let [r (marshall (.invoke m obj no-arg))]
           (if-not (nil? r)
             (hash-map
@@ -717,11 +747,12 @@
         obj)))
 
 (defn- use-aws-request-bean?
-  [method args]
+  [^Method method args]
   (let [types (.getParameterTypes method)]
-    (and (or (map? args) (< 1 (count args)))
+    (and (or (map? args) (< 1 (count args)) (nil? args))
          (< 0 (count types))
-         (or (map? args)
+         (or (nil? args)
+             (map? args)
              (and
                 (or (and
                       (even? (count args))
@@ -735,35 +766,36 @@
                   (not (< (count types) (count args))))))))
 
 (defn- prepare-args
-  [method args]
+  [^Method method args]
   (let [types (.getParameterTypes method)
         num   (count types)]
     (if (and (empty? args) (= 0 num))
       (into-array Object args)
       (if (= num (count args))
         (into-array Object
-          (map #(if (and (aws-package? %2) (seq (.getConstructors %2)))
-                  ; must be a concrete, instantiatable class
-                  (if (contains? @coercions %2)
-                      (coerce-value % %2)
-                      (create-bean %2 %))
-                  (coerce-value % %2))
-               (vec args)
-               types))
+                    (map (fn [arg ^Class clazz]
+                           (if (and (aws-package? clazz) (seq (.getConstructors clazz)))
+                             ; must be a concrete, instantiatable class
+                             (if (contains? @coercions clazz)
+                               (coerce-value arg clazz)
+                               (create-bean clazz arg))
+                             (coerce-value arg clazz)))
+                         (vec args)
+                         types))
         (if (use-aws-request-bean? method args)
           (cond
             (= 1 num)
             (into-array Object
                         [(create-request-bean
-                            method
-                            (seq (apply hash-map args)))])
+                           method
+                           (seq (apply hash-map args)))])
             (and (aws-package? (first types))
                  (= 2 num)
                  (= File (last types)))
             (into-array Object
                         [(create-request-bean
-                            method
-                            (seq (apply hash-map (butlast args))))
+                           method
+                           (seq (apply hash-map (butlast args))))
                          (last args)])))))))
 
 
@@ -783,7 +815,9 @@
           (instance? AWSCredentialsProvider (first args))
           (instance? AWSCredentials (first args)))
       {:args (if (-> args rest first map?)
-                 (mapcat identity (-> args rest args-from :args))
+                 (if (-> args rest first empty?)
+                     {}
+                     (mapcat identity (-> args rest args-from :args)))
                  (rest args))
        :credential (if (map? (first args))
                        (dissoc (first args) :client-config)
@@ -809,7 +843,7 @@
 (swap! client-config assoc :transfer-manager-fn (memoize transfer-manager*))
 
 (defn candidate-client
-  [clazz args]
+  [^Class clazz args]
   (let [cred-bound (or *credentials* (:credential args))
         credential (if (map? cred-bound)
                              (merge @credential cred-bound)
@@ -834,7 +868,7 @@
   "Returns a function that reflectively invokes method on
    clazz with supplied args (if any). The 'method' here is
    the Java method on the Amazon*Client class."
-  [clazz method & arg]
+  [clazz ^Method method & arg]
   (binding [*client-class* clazz]
     (let [args    (args-from arg)
           arg-arr (prepare-args method (:args args))
@@ -853,7 +887,7 @@
             (throw (.getTargetException ite))))))))
 
 (defn- types-match-args
-  [args method]
+  [args ^Method method]
   (let [types (.getParameterTypes method)]
     (if (and (= (count types) (count args))
              (every? identity (map instance? types args)))
@@ -868,18 +902,18 @@
       (first possible)
       (first
         (sort-by
-          (fn [method]
+          (fn [^Method method]
             (let [types (.getParameterTypes method)]
               (cond
                 (some coercible? types) 1
-                (some #(= java.lang.Enum (.getSuperclass %)) types) 2
+                (some #(= java.lang.Enum (.getSuperclass ^Class %)) types) 2
                 :else 3)))
           possible))))
 
 (defn possible-methods
   [methods args]
   (filter
-    (fn [method]
+    (fn [^Method method]
       (let [types (.getParameterTypes method)
             num   (count types)]
         (if (or
@@ -901,36 +935,150 @@
     (or (some (partial types-match-args args) methods)
         (choose-from (possible-methods methods args)))))
 
+(defn- clojure-case
+  "Similar to \"kabob case\" but the returned string is suitable for
+  reading a single symbol with `read-string`."
+  [string]
+  (-> string
+      ;; Replace the space between a non-upper-case letter and an
+      ;; upper-case letter with a dash.
+      (str/replace #"(?<=[^A-Z])(?=[A-Z])" "-")
+      ;; Remove anything that a Clojure reader would not accept in a
+      ;; symbol.
+      (str/replace #"[\\'\"\[\]\(\){}\s]" "")
+      (str/lower-case)))
+
+(defn- type-clojure-name
+  "Given a `java.lang.Class` return it's name in kabob case"
+  [^Class type]
+  (let [type-name (last (.. type getName (split "\\.")))
+        type-name (if-let [;; Check for a type name like "[C" etc.
+                           [_ name] (re-matches #"\[([A-Z]+)$" type-name)]
+                    name
+                    type-name)]
+    (clojure-case type-name)))
+
+(defn- parameter-clojure-name
+  "Given a `java.lang.reflect.Parameter` return it's name in kabob
+  case."
+  [^Parameter parameter]
+  (if (. parameter isNamePresent)
+    (clojure-case (. parameter getName))
+    ;; The name will be synthesized so instead we'll derive
+    ;; it from it's type.
+    (type-clojure-name (. parameter getType))))
+
+(def ^{:arglists '([method])
+       :private true}
+  parameter-names
+  "Given a `java.lang.reflect.Method` return a list of it's parameter
+  names."
+  ;; The regular expression here will only match against version
+  ;; numbers that are 1.7.X and below.
+  (if (re-matches #"[^2-9]\.[1-7]\..+" (System/getProperty "java.version"))
+    ;; Java 1.7 and below.
+    (fn [^Method method]
+      (map type-clojure-name (. method getParameterTypes)))
+    ;; Java 1.8 and above.
+    (fn [^Method method]
+      (map parameter-clojure-name (. method getParameters)))))
+
+(defn- method-arglist
+  "Derives a Clojure `:arglist` vector from a
+  `java.lang.reflect.Method`."
+  [method]
+  (let [names (parameter-names method)
+        type (some-> method .getParameters first .getType)
+        fields (when (class? type)
+                 (try
+                   (->> type
+                        .newInstance
+                        bean
+                        (remove (comp #{:class} first))
+                        (into {})
+                        (keys)
+                        (map (comp camel->keyword2 name))
+                        (map (comp symbol name))
+                        ;; fields common to all requests (calculated via `set/intersection`)
+                        (remove '#{clone-source request-credentials-provider request-metric-collector clone-root custom-request-headers sdk-client-execution-timeout request-credentials sdk-request-timeout custom-query-parameters request-client-options read-limit general-progress-listener})
+                        (sort)
+                        (vec))
+                   (catch InstantiationException _)))
+        ;; This will help determine when parameter names should be
+        ;; suffixed with an index i.e. `parameter-1`. Suffixing is
+        ;; necessary when parameter names are synthesized from their
+        ;; type names and the likelihood duplicates is high.
+        name-frequency (frequencies names)]
+    (loop [names names
+           ;; This map keeps track of the index of names when they
+           ;; appear more than once.
+           name-index {}
+           arglist []]
+      (if (empty? names)
+        (cond
+          (empty? arglist)
+          arglist
+
+          (= 1 (count arglist))
+          ['& (if fields
+                (assoc {:keys fields} ;; awkward construction for keeping a nice order when querying the arglist interactively (:keys will show up first this way)
+                       :as            (first arglist))
+                arglist)]
+
+          true
+          arglist)
+
+        (let [[name & names*] names]
+          (if (= (name-frequency name) 1)
+            (let [arg-symbol (symbol name)
+                  arglist* (conj arglist arg-symbol)]
+              (recur names*
+                     name-index
+                     arglist*))
+            ;; The parameter name appears more than once so we need to
+            ;; attach an index to it and update our name-index for the
+            ;; next parameter with the same name.
+            (let [index (get name-index name 1)
+                  name-index* (assoc name-index name (inc index))
+                  arg-symbol (symbol (str name "-" index))
+                  arglist* (conj arglist arg-symbol)]
+              (recur names*
+                     name-index*
+                     arglist*))))))))
+
 (defn intern-function
   "Interns into ns, the symbol mapped to a Clojure function
    derived from the java.lang.reflect.Method(s). Overloaded
    methods will yield a variadic Clojure function."
   [client ns fname methods]
-  (intern ns (with-meta (symbol (name fname))
-               {:amazonica/client client
-                :amazonica/methods methods})
-    (fn [& args]
-      (if-let [method (best-method methods args)]
-        (if-not args
-          ((fn-call client method))
-          ((fn-call client method args)))
-        (throw (IllegalArgumentException.
-                 (format "Could not determine best method to invoke for %s using arguments %s"
-                         (name fname) args)))))))
+  (let [source (some-> client pr-str munge (str/replace "." "/") (str ".java") (io/resource) str)]
+    (intern ns (with-meta (symbol (name fname))
+                 (cond-> {:amazonica/client  client
+                          :amazonica/methods methods
+                          :amazonica/method-name (-> methods first .getName)
+                          :arglists          (sort-by pr-str (map method-arglist methods))}
+                   source (assoc :amazonica/source source)))
+            (fn [& args]
+              (if-let [method (best-method methods args)]
+                (if-not args
+                  ((fn-call client method))
+                  ((fn-call client method args)))
+                (throw (IllegalArgumentException.
+                        (format "Could not determine best method to invoke for %s using arguments %s"
+                                (name fname) args))))))))
 
 (defn- client-methods
   "Returns a map with keys of idiomatic Clojure hyphenated keywords
   corresponding to the public Java method names of the class argument, vals are
   vectors of java.lang.reflect.Methods."
-  [client]
+  [^Class client]
   (->> (.getDeclaredMethods client)
-       (remove (fn [method]
+       (remove (fn [^Method method]
                  (let [mods (.getModifiers method)]
-                   (or (.isSynthetic method)
-                       (Modifier/isPrivate mods)
-                       (Modifier/isProtected mods)
-                       (Modifier/isStatic mods)))))
-       (group-by #(camel->keyword (.getName %)))))
+                   (or (not (Modifier/isPublic mods))
+                       (Modifier/isStatic mods)
+                       (.isSynthetic method)))))
+       (group-by #(camel->keyword (.getName ^Method %)))))
 
 (defn- show-functions [ns]
   (intern ns (symbol "show-functions")
@@ -947,7 +1095,7 @@
   (intern ns 'client-class client)
   (doseq [[fname methods] (client-methods client)
           :let [the-var (intern-function client ns fname methods)
-                fname2 (-> methods first .getName camel->keyword2)]]
+                fname2 (-> methods ^Method first .getName camel->keyword2)]]
     (when (not= fname fname2)
       (let [the-var2 (intern-function client ns fname2 methods)]
         (alter-meta! the-var assoc :amazonica/deprecated-in-favor-of the-var2)))))
