@@ -5,30 +5,34 @@
             [riemann.test :as test])
   (:use [riemann.common        :only [event]]
         [riemann.core          :only [stream!]]
-        [riemann.service       :only [Service ServiceEquiv]]
-        [clojure.tools.logging :only [info error]]))
+        [riemann.service       :only [Service ServiceEquiv ServiceStatus]]
+        [clojure.tools.logging :only [debug info error]]))
 
 (defn kafka
   "Returns a function that is invoked with a topic name and an optional message key and returns a stream. That stream is a function which takes an event or a sequence of events and sends them to Kafka.
-  
+
+  ```clojure
   (def kafka-output (kafka))
 
   (changed :state
     (kafka-output \"mytopic\"))
+  ```
 
   Options:
 
-  For a complete list of producer configuration options see https://kafka.apache.org/documentation/#producerconfigs  
+  For a complete list of producer configuration options see https://kafka.apache.org/documentation/#producerconfigs
 
-  :bootstrap.servers  Bootstrap configuration, default is \"localhost:9092\".
-  :value.serializer   Value serializer, default is json-serializer.
+  - :bootstrap.servers  Bootstrap configuration, default is \"localhost:9092\".
+  - :value.serializer   Value serializer, default is json-serializer.
 
   Example with SSL enabled:
 
+  ```clojure
   (def kafka-output (kafka {:bootstrap.servers \"kafka.example.com:9092\"
                             :security.protocol \"SSL\"
                             :ssl.truststore.location \"/path/to/my/truststore.jks\"
-                            :ssl.truststore.password \"mypassword\"}))"
+                            :ssl.truststore.password \"mypassword\"}))
+  ```"
 
   ([] (kafka {}))
   ([opts]
@@ -73,16 +77,21 @@
       (try
         (info "Subscribing to " topics "...")
         (client/subscribe! consumer topics)
-        (while running?
+        (while @running?
           (let [msgs (client/poll! consumer (:poll.timeout.ms opts))
                 msgs-by-topic (get msgs :by-topic)]
             (doseq [records msgs-by-topic
                     record (last records)]
-              (let [event (event (get record :value))]
-                (stream! @core event)))))
+              (try
+                (let [event (event (get record :value))]
+                  (stream! @core event))
+                (catch java.lang.NullPointerException e
+                  (debug (str "Invalid message: " record)))))))
         (catch Exception e
-          (error e "Interrupted consumption"))
-        (finally 
+          (do
+            (reset! running? false)
+            (error e "Interrupted consumption")))
+        (finally
           (client/close! consumer))))))
 
 (defn kafka-consumer
@@ -100,6 +109,8 @@
       ServiceEquiv
       (equiv? [this other]
         (= opts (:opts other)))
+      ServiceStatus
+      (running? [this] @running?)
       Service
       (conflict? [this other]
         (= opts (:opts other)))
