@@ -44,6 +44,10 @@
   (equiv? [s1 s2]
     (nil? s2)))
 
+(defprotocol ServiceStatus
+  (running? [service]
+          "Queries the running state of the service."))
+
 (defrecord ScheduledTaskService [name equiv-key interval delay f core task]
   ServiceEquiv
   (equiv? [this other]
@@ -136,18 +140,22 @@
 (defmacro all-equal?
   "Takes two objects to compare and a list of forms to compare them by.
 
+  ```clojure
   (all-equiv? foo bar
     (class)
     (foo 2)
     (.getSize))
+  ```
 
   becomes
 
+  ```clojure
   (let [a foo
         b bar]
     (and (= (class a) (class b))
          (= (foo 2 a) (foo 2 b))
-         (= (.getSize a) (.getSize b))))"
+         (= (.getSize a) (.getSize b))))
+  ```"
   [a b & forms]
   (let [asym (gensym "a__")
         bsym (gensym "b__")]
@@ -308,8 +316,10 @@
   Otherwise, services are equivalent when their class, name, and equiv-key are
   equal.
 
+  ```clojure
   (executor-service* :graphite {foo: 4}
-    #(ThreadPoolExecutor. 2 ...))"
+    #(ThreadPoolExecutor. 2 ...))
+  ```"
   ([name f] (executor-service name nil f))
   ([name equiv-key f]
    (ExecutorServiceService. name equiv-key f nil (atom nil))))
@@ -320,9 +330,9 @@
   any variables or function calls, they will be compared as code, not as
   their evaluated values.
 
-  OK:  (literal-executor-service :io (ThreadPoolExecutor. 2 ...))
-  OK:  (literal-executor-service :io (ThreadPoolExecutor. (inc 1) ...))
-  BAD: (literal-executor-service :io (ThreadPoolExecutor. x ...))"
+  - OK:  (literal-executor-service :io (ThreadPoolExecutor. 2 ...))
+  - OK:  (literal-executor-service :io (ThreadPoolExecutor. (inc 1) ...))
+  - BAD: (literal-executor-service :io (ThreadPoolExecutor. x ...))"
   [name executor-service-expr]
   `(executor-service
      ~name
@@ -333,11 +343,11 @@
   "An ExecutorServiceService based on a ThreadPoolExecutor with core and
   maximum threadpool sizes, and a LinkedBlockingQueue of a given size. Options:
 
-  :core-pool-size             Default 1
-  :max-pool-size              Default 128
-  :keep-alive-time            Default 10
-  :keep-alive-unit            Default MILLISECONDS
-  :queue-size                 Default 1000"
+  - :core-pool-size             Default 1
+  - :max-pool-size              Default 128
+  - :keep-alive-time            Default 10
+  - :keep-alive-unit            Default MILLISECONDS
+  - :queue-size                 Default 1000"
   ([name] (threadpool-service name {}))
   ([name {:keys [core-pool-size
                  max-pool-size
